{{
Cordic Resolve by Graham Stabler  19th Oct 06

With thanks to Chip Gracey for code it was based on and for excellent optimized *K code.

*
* This code isn't really an object as such, it just computes R.sin and R.cos and stops and is to be used.
* Together with the demo program it just proves the cordic code works.
* To use insert the code into your own asm program, you may want to remove the wrlong commands if the values
* are used within your asm program rather than passed out. 
*
* A full object will follow to allow a spin program to do arctan's and Sin/cos but many applications will just
* want to insert this assembly into another asm program
*

Cordic algorithm to compute R.cos(theta) and R.sin(theta).  You supply R as a signed 32 bit integer and
theta as an unsigned 32 bit integer that represents 0-360/0.

Remember an R.cos or R.sin will always produce a number that is some fraction of R, as the result is an integer
be sure to make R large enough to get some resolution.  The algorithm does the best it can by making small
numbers large before the cordic but for example if you do 100.Sin(45) you get 70 where as 1000.sin(45)
gives you 707 which could represent 70.7 etc etc  So in general try to think about your application and
the biggest number you will probably represent and scale so that is close to $EFFFFFFF (biggest positive
number that can be held).

A few reminders  
                                                      some common angles for testing with 
              /|                                      
             / |                                      45  = $20000000    /
         R  /  |                                      90  = $40000000    | 
           /   | R.Sin(theta) = Cy                    135 = $60000000    \ 
          /    |                                      180 = $80000000   <-  
         /)____|                                      225 = $A0000000    /  
             \                                        270 = $C0000000    |
               R.Cos(theta) = Cx                      315 = $E0000000    \

 }}

VAR

long    R_cos         ' The results held in hub Ram
long    R_sin
 

PUB start(theta,R) 
 _R := R             ' These two variables get loaded into the cog on the cognew
 ca := theta         ' because they are initialized variables
 cognew(@entry,@R_cos)
  

PUB get_cos : c
'' Get encoder count

  c := R_cos         ' Read from hub Ram

PUB get_sin : c
'' Get encoder count

  c := R_sin         ' Read from hub Ram

  
DAT                     'cx and cy will hold R.Cos(theta) and R.Sin(theta) respectively.
                        'ca contains input theta
                        '_R contains input R

                        'Angles will be input in the range of 0-360 the cordic algorthm will only output angles between +/-90 degrees.
                        'a combination of shifting the angle into the +/-90 range and conditional negation of the final answer fixes this.
                        '  _________________________________________________________
                        ' | Input angle range of theta | Target Angle | Final output | sign
                        ' |----------------------------------------------------------|
                        ' |          0 - 90            |   Theta      |    cx,  cy   |   0
                        ' |----------------------------------------------------------|
                        ' |         90 - 180           | Theta - 180  |   -cx, -cy   |   1
                        ' |----------------------------------------------------------|
                        ' |        180 - 270           | Theta - 180  |   -cx, -cy   |   1
                        ' |----------------------------------------------------------|
                        ' |        270 - 360           | Theta - 360  |    cx, -cy   |   2
                        '  ----------------------------------------------------------

                        org
entry                   mov     sign,#0                 ' sign of ouput (see table above)
                        movs    :lookup,#table          ' Puts address of table into sumc command
                        mov     t1,#0                   ' Initialize t1 
                        mov     t2,#20                  ' Initialize t2 
                        mov     ct,#0                   ' Initialize angle accumulator at 0, on x-axis
                        mov     shifts,#0
                        mov     _p,par

                       ' Firstly is is important to make sure that the input vector length is large,
                       ' dX and dY will become increasingly small as the the process ensues and as they
                       ' consist of the value of the previous Xi and Yi divided by a power of 2
                       ' these need to be big to begin with so they don't just go to zero (fractions are not possible) 

                       'Vector length is not signed but as loaded into Xi which is must not overflow into sine bit.

:makebig                mov     rt,_R
                        and     rt,_msb          wz
         if_z           shl     _R,#1
         if_z           add     shifts,#1        'record number of bit shifts
         if_z           jmp     #:makebig

                        ' Now multiply the input vector length R by K(0.607254)
                        ' Using Chip Gracey's ultra optimized 12 step plan.
                        ' Chip understands it at least :)
                        ' This make sure that the output is R.Cos(theta) and R.Sin(theta) and not scaled
                        ' by the gain of the Cordic process.

                        sar     _R,#1            'get %0.10001 of x
                        mov     t1,_R            '(      1    )
                        sar     t1,#4            '(+         1)
                        add     _R,t1

                        mov     t1,_R            'get %0.10011001 of x
                        sar     t1,#3            '(         10001)
                        add     t1,_R            '(+     10001   )

                        mov     t2,t1            'get %0.10011011011001 of x
                        sar     t2,#6            '(            10011001)
                        add     t2,t1            '(+     10011001      )

                        sar     _R,#11           'get %0.1001101101110101 of x
                        add     _R,t2            '(                 10001)
                                                 '(+     10011011011001  )

                        mov     t1,#0            ' re-initialize t1 (angle shift 2^-t)
                        mov     t2,#20           ' re-initialize t2 (number of interations)
                                                                     
                         
                        ' Alter theta and final sign to keep in Cordic's +/-90 range.
                        cmp     ca,a90          wc, wz  ' compare theta and 90
        if_nc_and_nz    sub     ca,a180                 ' if is greater than 90 subtract 180
        if_nc_and_nz    mov     sign,#1                 ' load negative flag 
        if_nc_and_nz    cmps    ca,a90          wc, wz  ' compare theta and 90
        if_nc_and_nz    sub     ca,a180                 ' if is greater than 90 subtract 180
        if_nc_and_nz    add     sign,#1                 ' clear sign flag          
                                                        ' Set estimate vector on x-axis
                        mov     cx,_R                   ' Initialize cx with K.R
                        mov     cy,#0                   ' Initialize cy with 0

                        'The loop
:loop                   'First calculate dX(i+1) => Yi.2^-t
                        mov     dx,cy                   ' Load dx with cy
                        sar     dx,t1                   ' Shifts dx by t1 while preserving sign
                        'Then  calculate  dY(i+1) => Xi.2^-t
                        mov     dy,cx                   ' Load dy with cx
                        sar     dy,t1                   ' Shifts dy by t1 while preserving sign

                        'Work out Di, i.e whether or not to add/subtract
                        
                        cmps    ct,ca           wc      ' c set if (estimate too small) 
                          
                        'Now add to the estimate with sign dependant on sign of Yi
                        sumc   cx,dx                    ' Add dx to cx if c=1 or -dx if c=0 
                        sumnc  cy,dy                    ' Add -dy to cx if c=1 or dy if c=0

                        'Next add or subract the specified rotation from the angle accumulator   
:lookup                 sumnc    ct,table                ' Add table to ca if c=1 or -table if c=0
                                                         ' Remembering that the source field of the sumc command "table"
                                                         ' is being over written to address different elements
                                                         ' of the table below
                                                         
                        add     :lookup,#1              ' Increments the destination of lookup sumc command
                        add     t1,#1                   ' increment t1, this is i
                        djnz    t2,#:loop               ' decrement t2, keep looping if not zero.

                        ' The results need to be scaled back
                        sar     cx,shifts
                        sar     cy,shifts
                        ' And signs changed to match
                        cmp     sign,#1         wz,wc
              if_z      neg     cx,cx                   'if sign 1 (both need negating)                 
              if_z      neg     cy,cy

                        wrlong  cx, _p                  'Write Xi to main memory. first plotted
                        add     _p,#4
                        wrlong  cy, _p                  'Write Yi to main memory  
:loop2
                        jmp     #:loop2                 ' And just loops here forever (dump this for proper programs)


'Table of atan(2^-i)

table                   long    $20000000        'atan(1)   = 45    degrees
                        long    $12E4051E        'atan(1/2) = 26.56 degrees
                        long    $09FB385B        'atan(1/4) 
                        long    $051111D4        'atan(1/8)
                        long    $028B0D43        'atan(1/16)
                        long    $0145D7E1        'atan(1/32)
                        long    $00A2F61E        'atan(1/64)
                        long    $00517C55        'atan(1/128)
                        long    $0028BE53        'atan(1/256)
                        long    $00145F2F        'atan(1/512)
                        long    $000A2F98        'atan(1/1024)
                        long    $000517CC        'atan(1/2048)
                        long    $00028BE6        'atan(1/4096)
                        long    $000145F3        'atan(1/8192)
                        long    $0000A2FA        'atan(1/16384)
                        long    $0000517D        'atan(1/32768)
                        long    $000028BE        'atan(1/65536)
                        long    $0000145F        'atan(1/131072)
                        long    $00000A30        'atan(1/262144)
                        long    $00000518        'atan(1/524288) = 1e-4 degrees

' Uninitialized data
_msb                    long    $40000000  
a180                    long    $80000000 '180 degrees
a90                     long    $40000000 '90  degrees
_K                      long    %10011011_01110100_11101101_10101000 ' is K, 0.607252935
ca                      long    0                
_R                      long    0

t1                      res     1
t2                      res     1
_p                      res     1
shifts                  res     1

dx                      res     1
dy                      res     1
cx                      res     1
cy                      res     1

ct                      res     1
rt                      res     1
sign                    res     1